/*
 * Copyright 2014-2015 Cloud Linux Zug GmbH
 *
 * Licensed under CLOUD LINUX LICENSE AGREEMENT
 * http://cloudlinux.com/docs/LICENSE.TXT
 *
 * The part the support library for lsapi & proxy_lsapi Apache modules
 * author Alexander Demeshko <ademeshko@cloudlinux.com>
 *
 */

#include <lscapi_util.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <unistd.h>

#include <http_request.h>
#include <util_script.h>

#ifdef APACHE2_2
#define ap_log_rerror_ ap_log_rerror
#endif

const int *lsapilog_module_index;

static void* alloc_by_request(size_t sz, void *user_data)
{
    return apr_palloc(((request_rec*)user_data)->pool, sz);
}

static const char get_request_body_key[] = "mod_lsapi-get_client_block-status";

static long get_request_body(char *buf, size_t bufLen, void *user_data )
{
    request_rec *r = (request_rec*)user_data;

    if(apr_table_get(r->notes, get_request_body_key)) {
        return 0;
    }

    //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "get_request_body(%lu) is called", *bufLenPtr);
    long readlen = ap_get_client_block(r, buf, bufLen);
    if(readlen < 0) {
        lscapi_rlog(LOG_ERR, 0, r, "get_client_block(%" APR_SIZE_T_FMT ") failed", bufLen);
        apr_table_set(r->notes, get_request_body_key, "failed");
    }
    //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "get_request_body(%lu): get_client_block read %lu", *bufLenPtr, readlen);
    else if(readlen == 0) {
        apr_table_set(r->notes, get_request_body_key, "eof");
    }
    return readlen;
}


void lscapi_log_error(const char *file, int line, int level, apr_status_t status,
                        const server_rec *s, const char *fmt, ...)
{
    char buf[8192];
    va_list args;

#ifndef APACHE2_2
    if (!APLOG_MODULE_IS_LEVEL(s,
                                 (lsapilog_module_index ? *lsapilog_module_index : APLOG_NO_MODULE),
                                 level))
    {
        return;
    }
#endif
    va_start(args, fmt);
    vsnprintf(buf, sizeof buf, fmt, args);

#ifdef APACHE2_2
    ap_log_error(file, line,
                    level, status, s, "[host %s] %s", s->server_hostname, buf);
#else
    ap_log_error_(file, line, (lsapilog_module_index ? *lsapilog_module_index : APLOG_NO_MODULE),
                    level, status, s, "[host %s] %s", s->server_hostname, buf);
#endif // APACHE2_2
    va_end(args);
}

void lscapi_log_rerror(const char *file, int line, int level,
                       apr_status_t status,
                        const request_rec *r, const char *fmt, ...)

{
    char buf[8192];
    va_list args;

#ifndef APACHE2_2
    if (!APLOG_R_MODULE_IS_LEVEL(r,
                                 (lsapilog_module_index ? *lsapilog_module_index : APLOG_NO_MODULE),
                                 level))
    {
        return;
    }
#endif
    va_start(args, fmt);
    vsnprintf(buf, sizeof buf, fmt, args);

#ifdef APACHE2_2
    ap_log_rerror(file, line,
                    level, status, r, "[host %s] %s", r->hostname, buf);
#else
    ap_log_rerror_(file, line, (lsapilog_module_index ? *lsapilog_module_index : APLOG_NO_MODULE),
                    level, status, r, "[host %s] %s", r->hostname, buf);
#endif // APACHE2_2
    va_end(args);
}

static uint8_t global_salt[LSCAPI_SALT_SIZE];

int lscapi_util_init(char *errbuf, size_t errlen)
{
#ifdef APACHE2_2
    apr_generate_random_bytes(global_salt, sizeof global_salt);
#else
    ap_random_insecure_bytes(global_salt, sizeof global_salt);
#endif

    return 0;
}


/*
apr_status_t lscapi_util_child_init(apr_pool_t *configpool, server_rec *s)
{
    //allocate hash uid/gid to socket name
    // register cleaner

    return APR_SUCCESS;
}
*/

/* in order to have random socket names */
#ifndef LSCAPI_WITH_RANDOM_SOCKET_NAMES
#define LSCAPI_WITH_RANDOM_SOCKET_NAMES
#endif


/*
#ifdef LSCAPI_WITH_RANDOM_SOCKET_NAMES
static int lscapi_get_socket_suffix(lscapi_rec *lscapi, uint32_t uid, uint32_t gid,
                                        request_rec *r, lsapi_svr_conf_t *svrcfg,
                                        char *_buf, size_t bufLen) {

    if(svrcfg->terminate_backends_on_exit) {

        if(bufLen < LSCAPI_ENCODED_SHA1_SIGN_WITHOUT_SALT_SIZE) {
            return -1;
        }

        uint32_t uidgid[2];
        uidgid[0] = uid;
        uidgid[1] = gid;
        uint8_t sign[LSCAPI_SHA1_SIGN_SIZE];

        lscapi_sign_uidgid_sha1(lscapi, uidgid, global_salt, sign);
        lscapi_encode_sha1_sign_without_salt(sign, _buf);

    } else { // if(svrcfg->terminate_backends_on_exit)

        char uidbuf[64];
        snprintf(uidbuf, sizeof uidbuf, "%u", uid);
        size_t uidLen = strlen(uidbuf);

        size_t len = strlen(r->handler);

        if(len + uidLen + 2 > bufLen) {
            return -1;
        }

        memcpy(_buf, r->handler, len);
        for(int i = 0; i < len; i++) {
            if(_buf[i] == '/') _buf[i] = '-';
        }
        _buf[len++] = '_';
        memcpy(_buf+len, uidbuf, uidLen+1);

    } // else of if(svrcfg->terminate_backends_on_exit)

    return 0;
}
#endif
*/

static apr_status_t check_script_file(request_rec *r, uint32_t uid, uint32_t gid,
                                      int check_document_root, int check_owner, 
                                      char *errbuf, size_t errlen)
{
    const char *script_filename = apr_table_get(r->subprocess_env, "SCRIPT_FILENAME");
    struct stat st_info;

    if(!script_filename) {
        snprintf(errbuf, errlen, "Could not determine script filename for request");
        return HTTP_SERVICE_UNAVAILABLE;
    }
    
    if(check_document_root)
    {
        const char *document_root;
#ifdef APACHE2_2
        document_root = ap_document_root(r);
#else
        document_root = ap_context_document_root(r);
#endif
        if (stat(document_root, &st_info) != 0) {
            snprintf(errbuf, errlen, "Could not stat document root (%s): errno %d", document_root, errno);
            return HTTP_SERVICE_UNAVAILABLE;
        }

        if(!(S_ISDIR(st_info.st_mode))) {
            snprintf(errbuf, errlen, "Document root (%s) is not directory", document_root);
            return HTTP_SERVICE_UNAVAILABLE;
        }

        if(st_info.st_uid != uid) {
            snprintf(errbuf, errlen, "document root (%s) is not owned by uid %u",
                    document_root, uid);
            return HTTP_SERVICE_UNAVAILABLE;
        }

        if (stat(script_filename, &st_info) != 0) {
            snprintf(errbuf, errlen, "Could not stat script filename (%s): errno %d", script_filename, errno);
            return HTTP_NOT_FOUND;
        }

        if(!(S_ISREG(st_info.st_mode))) {
        snprintf(errbuf, errlen, "Script file (%s) is not regular", script_filename);
        return HTTP_NOT_FOUND;
        }
    } else {//if(check_document_root)
        struct stat script_info;
        if (stat(script_filename, &script_info) != 0) {
            snprintf(errbuf, errlen, "Could not stat script filename (%s): errno %d", script_filename, errno);
            return HTTP_NOT_FOUND;
        }

        if(!(S_ISREG(script_info.st_mode))) {
            snprintf(errbuf, errlen, "Script file (%s) is not regular", script_filename);
            return HTTP_NOT_FOUND;
        }

    }
    
    if(check_owner) 
    {
        if(st_info.st_uid != uid) {
            snprintf(errbuf, errlen, "script file (%s) is not owned by uid %u",
                    script_filename, uid);
            return HTTP_SERVICE_UNAVAILABLE;
        }

        if( (st_info.st_mode & S_IWOTH) ) {
            snprintf(errbuf, errlen, "Script file (%s) is writable by others", script_filename);
            return HTTP_SERVICE_UNAVAILABLE;
        }
    } //if(check_owner)

    return OK;
}

static void set_auth(request_rec *r) {
    apr_table_unset(r->subprocess_env, "PHP_AUTH_USER");
    apr_table_unset(r->subprocess_env, "PHP_AUTH_PW");
    if(r->headers_in) {
        const char *auth = NULL;
        auth = apr_table_get(r->headers_in, "Authorization");
        if(auth && auth[0] != 0 && strncmp(auth, "Basic ", 6) == 0) {
            char *user;
            char *pass;
            user = ap_pbase64decode(r->pool, auth + 6);
            if(user) {
                pass = strchr(user, ':');
                if(pass) {
                    *pass++ = '\0';
                    char* auth_user = apr_pstrdup(r->pool, user);
                    char* auth_pass = apr_pstrdup(r->pool, pass);

                    apr_table_setn(r->subprocess_env, "PHP_AUTH_USER", auth_user);
                    apr_table_setn(r->subprocess_env, "PHP_AUTH_PW", auth_pass);
                }
            }
        }
    }
}


lscapi_rec *lscapi_create_connection(request_rec *r, const char *backend_path,
                                       unsigned *flagsPtr, apr_status_t *statusPtr,
                                       char *errbuf, size_t errlen,
                                       const module *m)
{
    lsapilog_module_index = &(m->module_index);

    apr_status_t status;
    char my_errbuf[128];
    ap_unix_identity_t *ugid = ap_run_get_suexec_identity(r);

    lsapi_dir_conf_t* cfg = (lsapi_dir_conf_t*) ap_get_module_config(r->per_dir_config, m);

    lsapi_svr_conf_t *svrcfg = ap_get_module_config(r->server->module_config, m);

    uint32_t lveuid;
    uint32_t lvegid;

#ifdef LSCAPI_DEBUG_UID_GID
    ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "uid nums: lsapi %d; suphp: %d; r: %d; itk: %d",
                  cfg->lsapi_uid_num, cfg->suphp_uid_num, cfg->ruid_num, cfg->itk_uid_num);
    for(int i = 0; i < cfg->lsapi_uid_num; i++) {
        ap_log_rerror(APLOG_MARK, LOG_ERR, 0, r, "lsapi uids %d/%d: %d/%d",
                      i, cfg->lsapi_uid_num, cfg->lsapi_uids[i], cfg->lsapi_gids[i] );

    }
    for(int i = 0; i < cfg->suphp_uid_num; i++) {
        ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "suphp uids %d/%d: %d/%d",
                      i, cfg->suphp_uid_num, cfg->suphp_uids[i], cfg->suphp_gids[i] );

    }
    for(int i = 0; i < cfg->ruid_num; i++) {
        ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "r uids %d/%d: %d/%d",
                      i, cfg->ruid_num, cfg->ruids[i], cfg->rgids[i] );

    }
    for(int i = 0; i < cfg->itk_uid_num; i++) {
        ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "lsapi uids %d/%d: %d/%d",
                      i, cfg->itk_uid_num, cfg->itk_uids[i], cfg->itk_gids[i] );

    }

    if(cfg->lsapi_uid_num > 0) {
        lveuid = cfg->lsapi_uids[cfg->lsapi_uid_num-1];
        lvegid = cfg->lsapi_gids[cfg->lsapi_uid_num-1];
        ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from lsapi", lveuid, lvegid);
    } else if(ugid) {
        lveuid = ugid->uid;
        lvegid = ugid->gid;
        ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from suexec", lveuid, lvegid);
    } else if(cfg->suphp_uid_num > 0) {
        lveuid = cfg->suphp_uids[cfg->suphp_uid_num-1];
        lvegid = cfg->suphp_gids[cfg->suphp_uid_num-1];
        ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from suphp", lveuid, lvegid);
    } else if(cfg->ruid_num > 0) {
        lveuid = cfg->ruids[cfg->ruid_num-1];
        lvegid = cfg->rgids[cfg->ruid_num-1];
        ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from ruid", lveuid, lvegid);
    } else if(cfg->itk_uid_num > 0) {
        lveuid = cfg->itk_uids[cfg->itk_uid_num-1];
        lvegid = cfg->itk_gids[cfg->itk_uid_num-1];
        ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from itk", lveuid, lvegid);
    } else if(svrcfg->use_default_uid) {
        lveuid = svrcfg->default_uid;
        lvegid = svrcfg->default_gid;
        ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from default", lveuid, lvegid);
    } else {
        ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid is absent");
        snprintf(errbuf, errlen, "Could not determine uid/gid for request");
        *statusPtr = HTTP_SERVICE_UNAVAILABLE;
        return NULL;
    }
#else
    if(cfg->lsapi_uid > 0) {
        lveuid = cfg->lsapi_uid;
        lvegid = cfg->lsapi_gid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from lsapi", lveuid, lvegid);
    } else if(ugid) {
        lveuid = ugid->uid;
        lvegid = ugid->gid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from suexec", lveuid, lvegid);
    } else if(cfg->suphp_uid > 0) {
        lveuid = cfg->suphp_uid;
        lvegid = cfg->suphp_gid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from suphp", lveuid, lvegid);
    } else if(cfg->ruid > 0) {
        lveuid = cfg->ruid;
        lvegid = cfg->rgid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from ruid", lveuid, lvegid);
    } else if(cfg->itk_uid > 0) {
        lveuid = cfg->itk_uid;
        lvegid = cfg->itk_gid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from itk", lveuid, lvegid);
    } else if(svrcfg->use_default_uid) {
        lveuid = svrcfg->default_uid;
        lvegid = svrcfg->default_gid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from default", lveuid, lvegid);
    } else {
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid is absent");
        snprintf(errbuf, errlen, "Could not determine uid/gid for request");
        *statusPtr = HTTP_SERVICE_UNAVAILABLE;
        return NULL;
    }
    if(cfg->lsapi_uid > 0) {
        lveuid = cfg->lsapi_uid;
        lvegid = cfg->lsapi_gid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from lsapi", lveuid, lvegid);
    } else if(ugid) {
        lveuid = ugid->uid;
        lvegid = ugid->gid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from suexec", lveuid, lvegid);
    } else if(cfg->suphp_uid > 0) {
        lveuid = cfg->suphp_uid;
        lvegid = cfg->suphp_gid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from suphp", lveuid, lvegid);
    } else if(cfg->ruid > 0) {
        lveuid = cfg->ruid;
        lvegid = cfg->rgid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from ruid", lveuid, lvegid);
    } else if(cfg->itk_uid > 0) {
        lveuid = cfg->itk_uid;
        lvegid = cfg->itk_gid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from itk", lveuid, lvegid);
    } else if(svrcfg->use_default_uid) {
        lveuid = svrcfg->default_uid;
        lvegid = svrcfg->default_gid;
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid %d/%d from default", lveuid, lvegid);
    } else {
        //ap_log_rerror(APLOG_MARK, LOG_DEBUG, 0, r, "ugid is absent");
        snprintf(errbuf, errlen, "Could not determine uid/gid for request");
        *statusPtr = HTTP_SERVICE_UNAVAILABLE;
        return NULL;
    }
#endif // LSCAPI_DEBUG_UID_GID

    ap_add_common_vars(r);
    ap_add_cgi_vars(r);
    apr_table_unset(r->subprocess_env, "GATEWAY_INTERFACE");

    set_auth(r);
    
    status = check_script_file(r, lveuid, lvegid, !svrcfg->skip_check_document_root, 
                               svrcfg->check_target_perm, errbuf, errlen);
    if(status != OK) {
        *statusPtr = status;
        return NULL;
    }

    lscapi_rec *lscapi = lscapi_create(alloc_by_request,
                                       (lscapi_log_fn)lscapi_log_rerror, r,
                                       lveuid, lvegid,
                                       flagsPtr,
                                       my_errbuf, sizeof my_errbuf);
    if(!lscapi) {
        snprintf(errbuf, errlen, "Could not create lsapi connection: %s", my_errbuf);
        *statusPtr = HTTP_SERVICE_UNAVAILABLE;
        return NULL;
    }

    lscapi_set_socket_path(lscapi, svrcfg->socket_path);

    lscapi_set_phprc(lscapi, svrcfg->phprc);

    lscapi_set_debug_enabled(lscapi, svrcfg->debug_enabled);

    lscapi_set_backend_path(lscapi, r->handler, backend_path);

    lscapi_set_user_spawn_backend(lscapi, (lscapi_spawn_backend_fn)lscapi_starter_send_spawn_cmd);

#ifdef APACHE2_2
    lscapi_set_random(lscapi, (lscapi_random_fn)apr_generate_random_bytes);
#else
    lscapi_set_random(lscapi, ap_random_insecure_bytes);
#endif

    lsapi_http_header_index_t *hdrIndexPtr;
    lsapi_header_offset_t *hdrOffsets;
    size_t hdrOffsetsNum;
    char *httpHeader;
    size_t httpHeaderLen;
    size_t contentLength;

    // Ignore Content-Length in the case of GET method
    // For example in the case of redirecting POST request with ErrorDocument
    if(r->method_number == M_GET) {
        apr_table_setn(r->headers_in, "Content-Length", "0");
    }

    lscapi_recreate_http_header(r, &hdrIndexPtr, &hdrOffsets,
                                &hdrOffsetsNum, &contentLength,
                                &httpHeader, &httpHeaderLen);

    // Ignore Content-Length in the case of Expect: 100-continue ?
/*
    if(contentLength) {
        const char *expect = apr_table_get(r->headers_in, "Expect");
        if(expect && !strcmp(expect, "100-continue")) {
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "Content-Length(%lu) is unreliable due to Expect:100-continue header", contentLength);
            contentLength = 0;
        }
    }
*/

    // set length of remaining request body and callback to read it
    lscapi_set_user_body_info(lscapi, contentLength, get_request_body,
                              (lscapi_should_get_block_fn)ap_should_client_block);

    lscapi_set_header_info(lscapi, hdrIndexPtr, hdrOffsets, hdrOffsetsNum,
                            httpHeader, httpHeaderLen);
    lscapi_var_t *envs;
    int envNum = lscapi_parse_server_vars(r, &envs);
    lscapi_set_envs(lscapi, envs, envNum);

    if(!cfg->mod_php_behaviour_off) {
        lscapi_var_t *specialEnvs;
        int specialEnvNum = lscapi_parse_special_vars(r, cfg->phpParams, &specialEnvs);
        lscapi_set_special_envs(lscapi, specialEnvs, specialEnvNum);
    }

    return lscapi;
}
