/*
 * Copyright 2014-2015 Cloud Linux Zug GmbH
 *
 * Licensed under CLOUD LINUX LICENSE AGREEMENT
 * http://cloudlinux.com/docs/LICENSE.TXT
 *
 * The part the support library for lsapi & proxy_lsapi Apache modules
 * author Alexander Demeshko <ademeshko@cloudlinux.com>
 *
 */

#include <lscapi_util.h>

#include <util_script.h>
#include <http_request.h>
#include <httpd.h>

struct hdrs_context {
    const lscapi_resphdr_info_t *hdrsInfoPtr;
    int curNum;
    int curHdrSize;
    apr_off_t curOff;
    request_rec *r;
    lsapi_svr_conf_t *svrcfg;
};

static int getsfunc_HDRBUF(char *w, int len, void *phdrCon) {
    struct hdrs_context *c = (struct hdrs_context*)phdrCon;

    if(c->curNum >= c->hdrsInfoPtr->hdrNum) { // normal exit
        w[0] = '\0';
        return 1;
    }

    if(c->curOff > c->hdrsInfoPtr->dataLen) {
        if(c->svrcfg->debug_enabled) {
            lscapi_rlog(APLOG_NOTICE, 0, c->r,
                        "Wrong packet format: offset %" APR_OFF_T_FMT " is more than datalen %"APR_SIZE_T_FMT,
                        c->curOff, c->hdrsInfoPtr->dataLen);
        }
        return -1;
    }

    if(!c->curHdrSize) {
        c->curHdrSize = c->hdrsInfoPtr->hdrSizes[c->curNum];
    }


    if(c->curHdrSize < len) {
        memcpy(w, c->hdrsInfoPtr->buf + c->curOff, c->curHdrSize);
        c->curNum++;
        c->curOff += c->curHdrSize;
        c->curHdrSize = 0;
    } else {
        memcpy(w, c->hdrsInfoPtr->buf + c->curOff, len-1);
        w[len-1] = '\0';
        c->curOff += len-1;
        c->curHdrSize -= len-1;
    }

    return 1;
}

#ifdef LSCAPI_WITH_DUMP_HEADER
/*
static void DumpHeaders(request_rec *r, const lscapi_resphdr_info_t *hdrsInfoPtr) {
    static unsigned long serialNo;
    char fname[256];


    snprintf(fname, sizeof fname, "/tmp/%s-%08lu.log", r->hostname, serialNo);
    serialNo++;

    FILE *fptr = fopen(fname, "wb");
    if(fptr) {
        lscapi_rlog(APLOG_ERR, 0, r,
                    "Dump headers into (%s)", fname);
        fwrite(hdrsInfoPtr->buf, 1, hdrsInfoPtr->dataLen, fptr);
        fclose(fptr);
    }
}
*/
#endif


// As RESPONSE_CODES in Apache 2.4 but should be the same also for Apache 2.2
#define LSAPI_RESPONSE_CODES 83


// On the base of the code from http_protocol.c from Apache 2.4
static const char * const status_lines[LSAPI_RESPONSE_CODES] =
{
    "100 Continue",
    "101 Switching Protocols",
    "102 Processing",
#define LEVEL_200  3
    "200 OK",
    "201 Created",
    "202 Accepted",
    "203 Non-Authoritative Information",
    "204 No Content",
    "205 Reset Content",
    "206 Partial Content",
    "207 Multi-Status",
    "208 Already Reported",
    NULL, /* 209 */
    NULL, /* 210 */
    NULL, /* 211 */
    NULL, /* 212 */
    NULL, /* 213 */
    NULL, /* 214 */
    NULL, /* 215 */
    NULL, /* 216 */
    NULL, /* 217 */
    NULL, /* 218 */
    NULL, /* 219 */
    NULL, /* 220 */
    NULL, /* 221 */
    NULL, /* 222 */
    NULL, /* 223 */
    NULL, /* 224 */
    NULL, /* 225 */
    "226 IM Used",
#define LEVEL_300 30
    "300 Multiple Choices",
    "301 Moved Permanently",
    "302 Moved Temporarily",    // instead of Apache "302 Found",
    "303 See Other",
    "304 Not Modified",
    "305 Use Proxy",
    NULL, /* 306 */
    "307 Temporary Redirect",
    "308 Permanent Redirect",
#define LEVEL_400 39
    "400 Bad Request",
    "401 Unauthorized",
    "402 Payment Required",
    "403 Forbidden",
    "404 Not Found",
    "405 Method Not Allowed",
    "406 Not Acceptable",
    "407 Proxy Authentication Required",
    "408 Request Timeout",
    "409 Conflict",
    "410 Gone",
    "411 Length Required",
    "412 Precondition Failed",
    "413 Request Entity Too Large",
    "414 Request-URI Too Long",
    "415 Unsupported Media Type",
    "416 Requested Range Not Satisfiable",
    "417 Expectation Failed",
    NULL, /* 418 */
    NULL, /* 419 */
    NULL, /* 420 */
    NULL, /* 421 */
    "422 Unprocessable Entity",
    "423 Locked",
    "424 Failed Dependency",
    NULL, /* 425 */
    "426 Upgrade Required",
    NULL, /* 427 */
    "428 Precondition Required",
    "429 Too Many Requests",
    NULL, /* 430 */
    "431 Request Header Fields Too Large",
#define LEVEL_500 71
    "500 Internal Server Error",
    "501 Not Implemented",
    "502 Bad Gateway",
    "503 Service Unavailable",
    "504 Gateway Timeout",
    "505 HTTP Version Not Supported",
    "506 Variant Also Negotiates",
    "507 Insufficient Storage",
    "508 Loop Detected",
    NULL, /* 509 */
    "510 Not Extended",
    "511 Network Authentication Required"
};


// got it from http_protocol.c from Apache 2.4
static int lsapi_index_of_response(int status)
{
    static int shortcut[6] = {0, LEVEL_200, LEVEL_300, LEVEL_400,
    LEVEL_500, LSAPI_RESPONSE_CODES};
    int i, pos;

    if (status < 100) {               /* Below 100 is illegal for HTTP status */
        return LEVEL_500;
    }

    for (i = 0; i < 5; i++) {
        status -= 100;
        if (status < 100) {
            pos = (status + shortcut[i]);
            if (pos < shortcut[i + 1] && status_lines[pos] != NULL) {
                return pos;
            }
            else {
                return LEVEL_500;            /* status unknown (falls in gap) */
            }
        }
    }
    return LEVEL_500;                         /* 600 or above is also illegal */
}

// got it from http_protocol.c from Apache 2.4
static const char* lsapi_get_status_line(int status)
{
    return status_lines[lsapi_index_of_response(status)];
}

#define SET_STATUS_WITH_LINE(r, st)  do { (r)->status = (st); (r)->status_line = lsapi_get_status_line((r)->status); } while(0)

static int ParseHeaders(request_rec *r, lsapi_svr_conf_t *svrcfg, const lscapi_resphdr_info_t *hdrsInfoPtr) {
    char buffer[MAX_STRING_LEN];
    struct hdrs_context context;
    context.hdrsInfoPtr = hdrsInfoPtr;
    context.curHdrSize = 0;
    context.curOff = sizeof(lsapi_resp_info) + sizeof(uint16_t)*hdrsInfoPtr->hdrNum;
    context.curNum = 0;
    context.svrcfg = svrcfg;
    context.r = r;

    int rc = ap_scan_script_header_err_core(r, buffer, getsfunc_HDRBUF, (void *) &context);
    if(hdrsInfoPtr->respStatus) SET_STATUS_WITH_LINE(r, hdrsInfoPtr->respStatus);
    return rc;
}

static int read_from_backend(lscapi_rec *lscapi, lsphp_conn_t *backend, request_rec *r,
                             lsapi_svr_conf_t *svrcfg, int drain, char *errbuf, size_t errlen) {
    int rc;
    char buf[LSAPI_MAX_DATA_PACKET_LEN];
    unsigned eventMask;
    apr_bucket_brigade *ob = apr_brigade_create(r->pool, r->connection->bucket_alloc);

    if(!drain) {
        ob = apr_brigade_create(r->pool, r->connection->bucket_alloc);
    }

    while(1) {

        size_t bufLen = LSAPI_MAX_DATA_PACKET_LEN;
        eventMask = 0;
        rc = lscapi_receive_response_chunk(backend, buf, &bufLen, &eventMask, errbuf, errlen);
        if(rc != 0) {
            lscapi_rlog(APLOG_ERR, 0, r, "Error receiving response: %s", errbuf);
            lscapi_set_error(lscapi);
            // TODO: it is too late to change status here, as header is already processed. But what we have to do?
            if(!drain) {
                SET_STATUS_WITH_LINE(r, HTTP_INTERNAL_SERVER_ERROR);
            }
            return -1;
        }

        if(eventMask & LSCAPI_BACKEND_LOG_RECEIVED) {
            if(eventMask & LSCAPI_BACKEND_LOG_FATAL) {
                lscapi_rlog(APLOG_ERR, 0, r, "Backend fatal error: %s", lscapi_get_backend_log(lscapi) );
            } else if(svrcfg->debug_enabled) {
                lscapi_rlog(APLOG_NOTICE, 0, r, "Backend log: %s", lscapi_get_backend_log(lscapi) );
            }
        }

        if(eventMask & LSCAPI_RESPONSE_FINISHED) {
            if(!drain) {
                apr_bucket *eosB = apr_bucket_eos_create(r->connection->bucket_alloc);
                APR_BRIGADE_INSERT_TAIL(ob, eosB);
                ap_pass_brigade(r->output_filters, ob);
            }
            break;
        }

        if(!drain) {
            apr_bucket *b = apr_bucket_transient_create(buf, bufLen, r->connection->bucket_alloc);
            APR_BRIGADE_INSERT_TAIL(ob, b);
            ap_pass_brigade(r->output_filters, ob);
        }

    } //while(!done)

    if(!drain) {
        //apr_brigade_cleanup(ib);
        apr_brigade_cleanup(ob);
    }
    return 0;
}

apr_status_t lscapi_do_request(lscapi_rec *lscapi, lsphp_conn_t *backend, request_rec *r, lsapi_svr_conf_t *svrcfg)
{
    char errbuf[256];
    int rc;
    unsigned eventMask;

    eventMask = 0;
    PROFILE_START(lscapi_send_request);
    rc = lscapi_send_request(backend, &eventMask, errbuf, sizeof errbuf);
    PROFILE_STOP(lscapi_send_request);
    if (rc != 0) {
        if(eventMask & LSCAPI_SENDREQ_INTERNAL_ERROR) {
            lscapi_rlog(APLOG_ERR, 0, r, "Internal error on sending request(%s); uri(%s) content-length(%s): %s",
                        r->the_request, r->unparsed_uri, apr_table_get(r->headers_in, "Content-Length"), errbuf);
            SET_STATUS_WITH_LINE(r, HTTP_INTERNAL_SERVER_ERROR);
        } else if(eventMask & LSCAPI_SENDREQ_BACKEND_ERROR) {
            lscapi_rlog(APLOG_ERR, 0, r, "Backend error on sending request(%s); uri(%s) content-length(%s) (lsphp is killed?): %s",
                        r->the_request, r->unparsed_uri, apr_table_get(r->headers_in, "Content-Length"), errbuf);
            SET_STATUS_WITH_LINE(r, HTTP_INTERNAL_SERVER_ERROR);
        } else if(eventMask & LSCAPI_SENDREQ_CLIENT_ERROR) {
            lscapi_rlog(APLOG_ERR, 0, r, "Client error on sending request(%s); uri(%s) content-length(%s): %s",
                        r->the_request, r->unparsed_uri, apr_table_get(r->headers_in, "Content-Length"), errbuf);
            SET_STATUS_WITH_LINE(r, HTTP_BAD_REQUEST);
        } else {
            // Unknown error?
            lscapi_rlog(APLOG_ERR, 0, r, "Error on sending request(%s); uri(%s) content-length(%s): %s",
                        r->the_request, r->unparsed_uri, apr_table_get(r->headers_in, "Content-Length"), errbuf);
            SET_STATUS_WITH_LINE(r, HTTP_INTERNAL_SERVER_ERROR);
        }
        lscapi_set_error(lscapi);
        return r->status;
    }

    if(eventMask & LSCAPI_BACKEND_LOG_RECEIVED) {
        if(eventMask & LSCAPI_BACKEND_LOG_FATAL) {
            lscapi_rlog(APLOG_ERR, 0, r, "Backend fatal error: %s", lscapi_get_backend_log(lscapi) );
        } else if(svrcfg->debug_enabled) {
            lscapi_rlog(APLOG_NOTICE, 0, r, "Backend log: %s", lscapi_get_backend_log(lscapi) );
        }
    }

    lscapi_resphdr_info_t hdrsInfo;

    eventMask = 0;
    PROFILE_START(lscapi_receive_response_header);
    rc = lscapi_receive_response_header(backend, &hdrsInfo, &eventMask, errbuf, sizeof errbuf);
    PROFILE_STOP(lscapi_receive_response_header);
    if(rc != 0) {
        lscapi_rlog(APLOG_ERR, 0, r, "Error receiving response header (lsphp is killed?): %s", errbuf);
        lscapi_set_error(lscapi);
        SET_STATUS_WITH_LINE(r, HTTP_INTERNAL_SERVER_ERROR);
        return r->status;
    }

    if(eventMask & LSCAPI_BACKEND_LOG_RECEIVED) {
        if(eventMask & LSCAPI_BACKEND_LOG_FATAL) {
            lscapi_rlog(APLOG_ERR, 0, r, "Backend fatal error: %s", lscapi_get_backend_log(lscapi) );
        } else if(svrcfg->debug_enabled) {
            lscapi_rlog(APLOG_NOTICE, 0, r, "Backend log: %s", lscapi_get_backend_log(lscapi) );
        }
    }

    rc = ParseHeaders(r, svrcfg, &hdrsInfo);
    if (rc == HTTP_NOT_MODIFIED)
    {
        return rc;
    } else if (rc != APR_SUCCESS)
    {
        read_from_backend(lscapi, backend, r, svrcfg, 1, errbuf, sizeof errbuf);  // discard output
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    const char *location = apr_table_get(r->headers_out, "Location");
    if (location && location[0] == '/' && r->status == 200)
    {
        read_from_backend(lscapi, backend, r, svrcfg, 1, errbuf, sizeof errbuf);  // discard output
        r->method = "GET";
        r->method_number = M_GET;
        apr_table_unset(r->headers_in, "Content-Length");

        ap_internal_redirect_handler(location, r);
        return OK;
    }
    else if (location && r->status == 200)
    {
        read_from_backend(lscapi, backend, r, svrcfg, 1, errbuf, sizeof errbuf);  // discard output
        return HTTP_MOVED_TEMPORARILY;
    }

    rc = read_from_backend(lscapi, backend, r, svrcfg, 0, errbuf, sizeof errbuf); // send output to browser through filters
    if(rc != 0) {
        return r->status;
    }

    return OK;
}
